
;**************************************************************************
;*                    I2C LCD Display demo programm                       *
;*                    v 2.06.s02 - 06.05.2013                             *
;**************************************************************************
;**************************************************************************
;*          Hardw.: PIC16F883 used                                        *
;*          OSC.......: Internal osc. used     POWER.....:   5V extern.   *
;*          LCD HD44780 compatible connected via I2C bus to the PIC       *
;*          duo LED connected to RB0 and RB1                              *
;**************************************************************************


;------------------------------------------------------
;device addresses:
;	PCF8574:		0100 A2 A1 A0 x
;	PCF8574A:		0111 A2 A1 A0 x
;------------------------------------------------------



 INCLUDE "P16F883.INC"


_U001			     EQU     B'1110000011000100'	;see Page 206 of Datasheet
_U002			     EQU     B'1111111111111111'	;see Page 207 of Datasheet

;	__CONFIG	_CONFIG1, _DEBUG_OFF & _PLL_EN & _BORV_2_5 & _BOR_OFF & _CP_OFF & _MCLR_DIS & _PWRT_DIS & _WDT_OFF & _INTRC_OSC_NOCLKOUT
;	__CONFIG	_CONFIG2, _VCAP_DIS


	__CONFIG	_CONFIG1, _U001

	__CONFIG	_CONFIG2, _U002


;==========================================================================
;       Variable Definition
;==========================================================================
TIMER1		EQU	H'20'		;delay routine
TIMER2		EQU	H'21'		; "	"	"

;interrupt
W_TEMP		EQU	H'22'
STATUS_TEMP	EQU	H'23'
PCLATH_TEMP	EQU	H'24'
data_tmp	EQU	H'25'

;----------------------------------------------------------------------------------
v_md_sbs	EQU	H'36'		;slave boards status:
					;          bit 7 = I2c display connected
;----------------------------------------------------------------------------------
v_pcf_add	EQU	H'45'		;PCF8574 I2C address

v_byte		EQU	H'5B'		;data to be written to PCF8574


;demo counter variables --------------------
v_cnt1		EQU	H'61'
v_cnt2		EQU	H'62'
v_cnt3		EQU	H'63'
v_cnt4		EQU	H'64'
v_cnt5		EQU	H'65'
v_cnt6		EQU	H'66'
v_cnt7		EQU	H'67'
v_cnt8		EQU	H'68'
v_cnt9		EQU	H'69'

v_disp_act	EQU	H'78'	;flag - I2C display active

;==========================================================================

		ORG	0		;Reset vector address
		GOTO	RESET		;goto RESET routine when boot.

;XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
;XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
;----------------------------------------------------------------------------
; INTERRUPT SERVICE ROUTINE -------------------------------------------------
;----------------------------------------------------------------------------
;XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
;XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

		ORG	004
;save important variables
		MOVWF	W_TEMP			;Copy W to W_TEMP register
		SWAPF	STATUS,W		;Swap status to be saved into W

;Swaps are used because they do not affect the status bits

		BANKSEL	STATUS_TEMP		;Select regardless of current bank
		MOVWF	STATUS_TEMP		;Copy status to bank zero STATUS_TEMP register
		MOVF	PCLATH,W		;Copy PCLATH to W register
		MOVWF	PCLATH_TEMP		;Copy W register to PCLATH_TEMP

		bcf	INTCON,GIE

		PAGESEL	PAGE0

		goto	int_serv
;----------------------------------------------------------------------------
;----------------------------------------------------------------------------
;
;	       **********************************
;              **  RESET :  main boot routine  **
;              **********************************

RESET		nop
PAGE0
;
		BSF	STATUS,RP0	;Switch to register bank 1
		MOVLW	B'01011111'	;Set PIC options (See datasheet).
		MOVWF	OPTION_REG	;Write the OPTION register.
		bcf	OPTION_REG,T0CS	;set-up TIMER0 as timer
		bsf	OPTION_REG,PSA	;prescaler NOT for TIMER0
		;
		CLRF	INTCON		;Disable interrupts
		clrf	IOCB
		bsf	IOCB,D'000'	;enable interrupt on change for RB0 only

		MOVLW	B'11111111'	;PORTA (not used here)
		MOVWF	TRISA

		MOVLW	B'11111100'	;RB0 and RB1 are outputs for the duo-LED
		MOVWF	TRISB

		MOVLW	B'11111111'	;PORTC (important is that RC4 and RC5 are configured as inputs for proper I2C communication)
		MOVWF	TRISC

		MOVLW	B'11111111'	;PORTE (not used)
		MOVWF	TRISE

		movlw	B'00000011'
		movwf	WPUB

		bsf	STATUS,RP0
		bsf	STATUS,RP1
		clrf	ANSEL
		clrf	ANSELH

		bcf	STATUS,RP0	;Switch Back to reg. Bank 0
		bcf	STATUS,RP1

;----------------------------------------------------------------------------
;set up I2C BUS for communication as a master
		movlw	B'00101000'		;see Datasheet Page 177
;			bit 3-0 SSPM<3:0>: Synchronous Serial Port Mode Select bits
;			1000 = I2C Master mode, clock = FOSC / (4 * (SSPADD+1))
		movwf	SSPCON

;----------------------------------------------------------------------------


; XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
; XXXXXXXXXX       initialization  XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
; XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX


		clrf	PIE1
		clrf	PIE2

		bsf	PIE1,SSPIE
;

		clrf	v_disp_act
		clrf	v_md_sbs

		movlw	'0'
		movwf	v_cnt1
		movwf	v_cnt2
		movwf	v_cnt3
		movwf	v_cnt4
		movwf	v_cnt5
		movwf	v_cnt6
		movwf	v_cnt7
		movwf	v_cnt8
		movwf	v_cnt9

;XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
;XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
;-----------------------------------------------------------------------------
;                          Main part
;-----------------------------------------------------------------------------
;XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
;XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
;

;switch the red LED on
m001		call	RED_ON
		bcf	v_disp_act,D'000'	;I2C display not active (anymore)
;check if I2C LCD is connected
m002		call	sb_check		;check slave boards
		call	w_i2c_init		;I2C display initialization (if required)
		call	inc_counter		;incerement demo counter

 		btfss	v_md_sbs,D'007'
		goto	m001

		btfsc	v_disp_act,D'000'
		goto	disp_active
;display is not active
		goto	m001
;----------------------------------------
;display is active - send the counter values to 2nd line
disp_active	movlw	B'11000000'
		call	send_comm_i2c
		movlw	'C'
		call	send_data_i2c
		movlw	'n'
		call	send_data_i2c
		movlw	't'
		call	send_data_i2c
		movlw	':'
		call	send_data_i2c
		movf	v_cnt1,0
		call	send_data_i2c
		movf	v_cnt2,0
		call	send_data_i2c
		movf	v_cnt3,0
		call	send_data_i2c
		movf	v_cnt4,0
		call	send_data_i2c
		movf	v_cnt5,0
		call	send_data_i2c
		movf	v_cnt6,0
		call	send_data_i2c
		movf	v_cnt7,0
		call	send_data_i2c
		movf	v_cnt8,0
		call	send_data_i2c
		movf	v_cnt9,0
		call	send_data_i2c

		goto	m002
;-----------------------------------------------------------------------------
RED_ON		bcf	PORTB,D'000'
		bsf	PORTB,D'001'
		return
;-----------------------------------------------------------------------------
GREEN_ON	bsf	PORTB,D'000'
		bcf	PORTB,D'001'
		return
;-----------------------------------------------------------------------------
LED_OFF		bcf	PORTB,D'000'
		bcf	PORTB,D'001'
		return
;-----------------------------------------------------------------------------
inc_counter	incf	v_cnt9,1
		movf	v_cnt9,0
		xorlw	':'
		btfss	STATUS,Z
		return
;
		movlw	'0'
		movwf	v_cnt9
		incf	v_cnt8,1
		movf	v_cnt8,0
		xorlw	':'
		btfss	STATUS,Z
		return
;
		movlw	'0'
		movwf	v_cnt8
		incf	v_cnt7,1
		movf	v_cnt7,0
		xorlw	':'
		btfss	STATUS,Z
		return
;
		movlw	'0'
		movwf	v_cnt7
		incf	v_cnt6,1
		movf	v_cnt6,0
		xorlw	':'
		btfss	STATUS,Z
		return
;
		movlw	'0'
		movwf	v_cnt6
		incf	v_cnt5,1
		movf	v_cnt5,0
		xorlw	':'
		btfss	STATUS,Z
		return
;
		movlw	'0'
		movwf	v_cnt5
		incf	v_cnt4,1
		movf	v_cnt4,0
		xorlw	':'
		btfss	STATUS,Z
		return
;
		movlw	'0'
		movwf	v_cnt4
		incf	v_cnt3,1
		movf	v_cnt3,0
		xorlw	':'
		btfss	STATUS,Z
		return
;
		movlw	'0'
		movwf	v_cnt3
		incf	v_cnt2,1
		movf	v_cnt2,0
		xorlw	':'
		btfss	STATUS,Z
		return
;
		movlw	'0'
		movwf	v_cnt2
		incf	v_cnt1,1
		movf	v_cnt1,0
		xorlw	':'
		btfss	STATUS,Z
		return
;
		movlw	'0'
		movwf	v_cnt1
;
		return
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
; Slave boards check - check, if an slave board is connected
;-----------------------------------------------------------------------------
;  I2C display board address: 0111000x
;
;if an slave board is active, we got always bit 3 in response eq to 0
;display board responses with bit 7 eq to 0
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
sb_check	nop
		clrf	v_md_sbs
;board #D check
		movlw	B'01110000'		;I2C address for I2C LCD
		movwf	v_pcf_add
		call	pcf_i2c_read

		btfss	v_byte,D'007'
		bsf	v_md_sbs,D'007'		;I2C is active

		return
;xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
;xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
;xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
;
; I2C communication routines
;
;xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
;xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
;xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
; PCF8574 I2C communication routines - WRITE
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------

pcf_i2c_write	bsf	STATUS,RP0
		movlw	B'00111111'
		movwf	SSPADD
		clrf	SSPSTAT
		bcf	STATUS,RP0

;2nd step - Start condition
		bsf	STATUS,RP0
		bsf	SSPCON2, SEN
		bcf	STATUS,RP0

		call	d55s
		bcf	PIR1, SSPIF
;
;Step 3: addressing the PCF8574 device
		movf	v_pcf_add,0
		andlw	B'11111110'		;we want to read - last bit needs to be "0"
		movwf	SSPBUF

		btfsc	SSPCON,WCOL
		nop

		bcf	PIR1, SSPIF
		call	d55s
		bcf	PIR1, SSPIF

;Step 4 - write data to the device
		movf	v_byte,0
		movwf	SSPBUF

		btfsc	SSPCON,WCOL
		nop

		bcf	PIR1, SSPIF
		call	d55s
		bcf	PIR1, SSPIF

		return

;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
; PCF8574 I2C communication routines - READ
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------

pcf_i2c_read	bsf	STATUS,RP0
		movlw	B'00111111'
		movwf	SSPADD
		clrf	SSPSTAT
		bcf	STATUS,RP0

;2nd step - Start condition
		bsf	STATUS,RP0
		bsf	SSPCON2, SEN
		bcf	STATUS,RP0

		call	d55s
		bcf	PIR1, SSPIF
;
;Step 3: addressing the PCF8574 device

		movf	v_pcf_add,0		;get the device address
		iorlw	B'00000001'		;we want to read - last bit needs to be "1"
		movwf	SSPBUF

		btfsc	SSPCON,WCOL
		nop

		bcf	PIR1, SSPIF
		call	d55s
		bcf	PIR1, SSPIF

;Step 4 - read data from the device
		bsf	STATUS,RP0
		bsf	SSPCON2, RCEN
		bcf	STATUS,RP0
		
		call	d55s
		bcf	PIR1, SSPIF

		movf	SSPBUF,0
		movwf	v_byte

		btfsc	SSPCON,WCOL
		nop

;------------------------------------------------------------------
; no ACK at this place --- it's killing communication ------------
;		bsf	STATUS,RP0
;		bcf	SSPCON2, ACKDT		;ACK
;		bsf	SSPCON2, ACKEN
;		bcf	STATUS,RP0
;		
;		call	d55
;		bcf	PIR1, SSPIF
;------------------------------------------------------------------
;Step 5: send STOP condition

		bsf	STATUS,RP0
		bsf	SSPCON2, PEN
		bcf	STATUS,RP0

		bcf	PIR1, SSPIF
		call	d55s

		return
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
;XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
;XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

;		*********************************************
;		*               delay routines              *
;		*********************************************

d55s		movlw	D'255'
		movwf	TIMER2
d55s_loop	decfsz	TIMER2,F
		goto	d55s_loop
		return
;-----------------------------------------------------------------------------
dy_comm		MOVLW   D'010'
		MOVWF   TIMER2
DEL_LOOP22c	DECFSZ  TIMER2,F
                GOTO    DEL_LOOP22c
		return
;-----------------------------------------------------------------------------
d2_p2		MOVLW   D'255'
		MOVWF   TIMER2
DEL_LOOPx1	MOVLW   D'255'
		MOVWF   TIMER1
DEL_LOOPx2	DECFSZ  TIMER1,F
		GOTO    DEL_LOOPx2
		DECFSZ  TIMER2,F
		GOTO    DEL_LOOPx1
		return
;---------------------------------------------------------------------------
dy_comm_i2c	MOVLW   D'002'
		MOVWF   TIMER2
DEL_LOOP22c2	DECFSZ  TIMER2,F
                GOTO    DEL_LOOP22c2
		return
;---------------------------------------------------------------------------
d1_i2c		MOVLW   D'020'
		MOVWF   TIMER2
DEL_LOOP11q	DECFSZ  TIMER2,F
                GOTO    DEL_LOOP11q
		return
;------------------------------------------------------------------------------
;-----------------------------------------------------------------------------

int_serv	nop
		nop
		nop

		bcf	PIR1, SSPIF

;restore values
		BANKSEL STATUS_TEMP	;Select regardless of current bank
		MOVF PCLATH_TEMP,W	;
		MOVWF PCLATH		;Restore PCLATH
		SWAPF STATUS_TEMP,W	;Swap STATUS_TEMP register into W
					;(sets bank to original state)
		MOVWF STATUS		;Move W into STATUS register
		SWAPF W_TEMP,F		;Swap W_TEMP
		SWAPF W_TEMP,W		;Swap W_TEMP into W


		bsf	INTCON,GIE
		bsf	INTCON,PEIE

		bsf	PIE1,SSPIE
		retfie

;xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
;x             I2C external display communication routines                                      x
;xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
w_i2c_init	nop
		btfss	v_md_sbs,D'007'
		return				;I2C display not connected
		btfsc	v_disp_act,D'000'
		return				;I2C display activated already
;---------------------------------------------------------------------------------
;now we have to do initialization
;------------------------------
;RS  = P4
;R/W = GND
;E   = P5

;D0 - D3 -- not connected

;D4  = P0
;D5  = P1
;D6  = P2
;D7  = P3
;------------------------------
		bsf	v_disp_act,D'000'	;initialzation (will be) done

;1st step of init:
		clrf	v_byte
		movlw	B'01110000'
		movwf	v_pcf_add		;I2C address for PCF8574 (I2C display)
		PAGESEL	PAGE0
		call	pcf_i2c_write		;#1
;
		bsf	v_byte,D'000'
		bsf	v_byte,D'001'
		bsf	v_byte,D'005'		;EN
		call	pcf_i2c_write
		call	dy_comm			;#2
		bcf	v_byte,D'005'
		call	pcf_i2c_write
		call	d2_p2			;#3

		bsf	v_byte,D'005'		;EN
		call	pcf_i2c_write
		call	dy_comm			;#4
		bcf	v_byte,D'005'
		call	pcf_i2c_write
		call	d2_p2			;#5

		bsf	v_byte,D'005'		;EN
		call	pcf_i2c_write
		call	dy_comm			;#6
		bcf	v_byte,D'005'
		call	pcf_i2c_write
		call	d2_p2			;#7

		bcf	v_byte,D'000'
		bsf	v_byte,D'005'		;EN
		call	pcf_i2c_write
		call	dy_comm			;#8
		bcf	v_byte,D'005'
		call	pcf_i2c_write
		call	d2_p2			;#9
;2nd step of init:
		movlw	B'00101000'
		call	send_comm_i2c
;3rd step of init:
		movlw	B'00001000'
		call	send_comm_i2c

		movlw	B'00001111'
		call	send_comm_i2c
;4th step of init:
		movlw	B'00000110'
		call	send_comm_i2c
;clear the display
		movlw	B'00000001'
		call	send_comm_i2c
;----------------------------------------------------------------------
		call	GREEN_ON
;-----------------------------------------------
		movlw	B'10000000'
		call	send_comm_i2c
		movlw	'T'
		call	send_data_i2c
		movlw	'e'
		call	send_data_i2c
		movlw	's'
		call	send_data_i2c
		movlw	't'
		call	send_data_i2c
		movlw	'.
		call	send_data_i2c
		movlw	'.
		call	send_data_i2c
		movlw	'.
		call	send_data_i2c
;-----------------------------------------------
		return

;-----------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------
send_comm_i2c	btfss	v_disp_act,D'000'
		return				;display not connected
		movwf	data_tmp

		swapf	data_tmp,0
		andlw	B'00001111'
		movwf	v_byte			;higher 4 bits are put to RA0-RA3
		call	scbc_i2c		;send

		movf	data_tmp,0
		andlw	B'00001111'
		movwf	v_byte			;lower 4 bits are put to RA0-RA3
		call	scbc_i2c		;send

		movf	data_tmp,0
		return

;set control bits
scbc_i2c	movlw	B'01110000'
		movwf	v_pcf_add		;I2C address for PCF8574 (I2C display)

		bcf	v_byte,D'004'		;RS -> 0
		bsf	v_byte,D'005'		;EN -> 1
		call	pcf_i2c_write		;#1
		call	dy_comm_i2c		;wait a while... -- just a short break for I2C

		bcf	v_byte,D'005'		;EN ->0
		call	pcf_i2c_write		;#1
		call	d1_i2c			;wait a little bit... -- just a short break for I2C
		return
;-----------------------------------------------------------------------------------------------
send_data_i2c	btfss	v_disp_act,D'000'
		return				;display not connected
		movwf	data_tmp

		swapf	data_tmp,0
		andlw	B'00001111'
		movwf	v_byte			;higher 4 bits are put to RA0-RA3
		call	scbd_i2c		;send

		movf	data_tmp,0
		andlw	B'00001111'
		movwf	v_byte			;lower 4 bits are put to RA0-RA3
		call	scbd_i2c		;send

		movf	data_tmp,0
		return

;set control bits
scbd_i2c	movlw	B'01110000'
		movwf	v_pcf_add		;I2C address for PCF8574 (I2C display)

		bsf	v_byte,D'004'		;RS -> 1
		bsf	v_byte,D'005'		;EN -> 1
		call	pcf_i2c_write		;#1
		call	dy_comm_i2c		;wait a while... -- just a short break for I2C

		bcf	v_byte,D'005'		;EN ->0
		call	pcf_i2c_write		;#1
		call	d1_i2c			;wait a little bit... -- just a short break for I2C
		return
;-----------------------------------------------------------------------------------------------
;xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
;XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
;XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX


		END